@@ -574,7 +574,6 @@ int add_configfs_node(int nodeid, char *addr, int addrlen, int local)
 {
 	char path[PATH_MAX];
 	char padded_addr[sizeof(struct sockaddr_storage)];
-	struct sockaddr_storage *saddr = (struct sockaddr_storage *)addr;
 	char buf[32];
 	int rv, fd;
 
@@ -619,31 +618,9 @@ int add_configfs_node(int nodeid, char *addr, int addrlen, int local)
 	/*
 	 * set the address
 	 */
+
 	memset(padded_addr, 0, sizeof(padded_addr));
-
-	if (local) {
-		if (saddr->ss_family == AF_INET && opt(enable_inaddr_any_ind) != 6) {
-			struct sockaddr_in *in4  = (struct sockaddr_in *)&padded_addr;
-
-			if (opt(enable_inaddr_any_ind))
-				in4->sin_addr.s_addr = INADDR_ANY;
-			else
-				memcpy(in4, addr, addrlen);
-
-			in4->sin_family = AF_INET;
-		} else {
-			struct sockaddr_in6 *in6  = (struct sockaddr_in6 *)padded_addr;
-			struct in6_addr in6addr_any = IN6ADDR_ANY_INIT;
-
-			if (opt(enable_inaddr_any_ind))
-				memcpy(in6->sin6_addr.s6_addr, &in6addr_any, sizeof(in6addr_any));
-			else
-				memcpy(in6, addr, addrlen);
-
-			in6->sin6_family = AF_INET6;
-		}
-	} else
-		memcpy(padded_addr, addr, addrlen);
+	memcpy(padded_addr, addr, addrlen);
 
 	memset(path, 0, PATH_MAX);
 	snprintf(path, PATH_MAX, "%s/%d/addr", COMMS_DIR, nodeid);
@@ -685,6 +662,25 @@ int add_configfs_node(int nodeid, char *addr, int addrlen, int local)
 		return -1;
 	}
 	close(fd);
+
+	if (opt(bind_all_ind)) {
+		memset(path, 0, PATH_MAX);
+		snprintf(path, PATH_MAX, "%s/%d/bind_all", COMMS_DIR, nodeid);
+
+		fd = open(path, O_WRONLY);
+		if (fd < 0) {
+			log_error("%s: open failed: %d", path, errno);
+			return -1;
+		}
+
+		rv = do_write(fd, (void *)"1", strlen("1"));
+		if (rv < 0) {
+			log_error("%s: write failed: %d", path, errno);
+			close(fd);
+			return -1;
+		}
+		close(fd);
+	}
  out:
 	return 0;
 }
