@@ -315,6 +315,24 @@ void dlm_send_message(struct lockspace *ls, char *buf, int len)
 	_send_message(ls->cpg_handle, buf, len, type);
 }
 
+int dlm_send_message_daemon(char *buf, int len)
+{
+	struct dlm_header *hd = (struct dlm_header *) buf;
+	int type = hd->type;
+
+	hd->version[0]  = cpu_to_le16(our_protocol.daemon_run[0]);
+	hd->version[1]  = cpu_to_le16(our_protocol.daemon_run[1]);
+	hd->version[2]  = cpu_to_le16(our_protocol.daemon_run[2]);
+	hd->type	= cpu_to_le16(hd->type);
+	hd->nodeid      = cpu_to_le32(our_nodeid);
+	hd->to_nodeid   = cpu_to_le32(hd->to_nodeid);
+	hd->flags       = cpu_to_le32(hd->flags);
+	hd->msgdata     = cpu_to_le32(hd->msgdata);
+	hd->msgdata2    = cpu_to_le32(hd->msgdata2);
+
+	return _send_message(cpg_handle_daemon, buf, len, type);
+}
+
 void dlm_header_in(struct dlm_header *hd)
 {
 	hd->version[0]  = le16_to_cpu(hd->version[0]);
@@ -329,6 +347,50 @@ void dlm_header_in(struct dlm_header *hd)
 	hd->msgdata2    = le32_to_cpu(hd->msgdata2);
 }
 
+static void run_info_out(struct run_info *info)
+{
+	info->dest_nodeid  = cpu_to_le32(info->dest_nodeid);
+	info->start_nodeid = cpu_to_le32(info->start_nodeid);
+	info->local_pid    = cpu_to_le32(info->local_pid);
+	info->local_result = cpu_to_le32(info->local_result);
+	info->need_replies = cpu_to_le32(info->need_replies);
+	info->reply_count  = cpu_to_le32(info->reply_count);
+	info->fail_count   = cpu_to_le32(info->fail_count);
+	info->flags        = cpu_to_le32(info->flags);
+}
+
+static void run_info_in(struct run_info *info)
+{
+	info->dest_nodeid  = le32_to_cpu(info->dest_nodeid);
+	info->start_nodeid = le32_to_cpu(info->start_nodeid);
+	info->local_pid    = le32_to_cpu(info->local_pid);
+	info->local_result = le32_to_cpu(info->local_result);
+	info->need_replies = le32_to_cpu(info->need_replies);
+	info->reply_count  = le32_to_cpu(info->reply_count);
+	info->fail_count   = le32_to_cpu(info->fail_count);
+	info->flags        = le32_to_cpu(info->flags);
+}
+
+static void run_request_out(struct run_request *req)
+{
+	run_info_out(&req->info);
+}
+
+static void run_request_in(struct run_request *req)
+{
+	run_info_in(&req->info);
+}
+
+static void run_reply_out(struct run_reply *rep)
+{
+	run_info_out(&rep->info);
+}
+
+static void run_reply_in(struct run_reply *rep)
+{
+	run_info_in(&rep->info);
+}
+
 int dlm_header_validate(struct dlm_header *hd, int nodeid)
 {
 	if (hd->version[0] != our_protocol.daemon_run[0] ||
@@ -1974,6 +2036,12 @@ static void deliver_cb_daemon(cpg_handle_t handle,
 	case DLM_MSG_FENCE_CLEAR:
 		receive_fence_clear(hd, len);
 		break;
+	case DLM_MSG_RUN_REQUEST:
+		receive_run_request(hd, len);
+		break;
+	case DLM_MSG_RUN_REPLY:
+		receive_run_reply(hd, len);
+		break;
 	default:
 		log_error("deliver_cb_daemon unknown msg type %d", hd->type);
 	}
@@ -1981,6 +2049,204 @@ static void deliver_cb_daemon(cpg_handle_t handle,
 	daemon_fence_work();
 }
 
+int receive_run_reply(struct dlm_header *hd, int len)
+{
+	struct run_reply *rep = (struct run_reply *)hd;
+	struct run *run;
+	int i;
+
+	run_reply_in(rep);
+
+	log_debug("receive_run_reply %s from %d result %d",
+		  rep->uuid, hd->nodeid, rep->info.local_result);
+
+	if (!opt(enable_helper_ind)) {
+		log_debug("receive_run_reply %s helper not enabled", rep->uuid);
+		return 0;
+	}
+
+	run = find_run(rep->uuid);
+	if (!run) {
+		log_debug("receive_run_reply no uuid %s", rep->uuid);
+		return 0;
+	}
+
+	/*
+	 * Only the starting node keeps track of results.
+	 */
+	if (run->info.start_nodeid != our_nodeid)
+		return 0;
+
+	if (len != sizeof(struct run_reply)) {
+		log_debug("receive_run_reply %s bad len %s expect %d",
+			  rep->uuid, len, sizeof(struct run_reply));
+		run->info.reply_count++;
+		run->info.need_replies--;
+		return 0;
+	}
+
+	for (i = 0; i < run->node_count; i++) {
+		if (run->node_results[i].nodeid != hd->nodeid)
+			continue;
+
+		/* shouldn't happen? */
+		if (run->node_results[i].replied)
+			break;
+
+		run->node_results[i].result = rep->info.local_result;
+		run->node_results[i].replied = 1;
+
+		if (rep->info.local_result)
+			run->info.fail_count++;
+
+		run->info.reply_count++;
+		run->info.need_replies--;
+
+		/*
+		log_debug("run reply_count % need_replies %d fail_count %d",
+			  run->info.reply_count, run->info.need_replies, run->info.fail_count);
+		*/
+		break;
+	}
+
+	return 0;
+}
+
+int receive_run_request(struct dlm_header *hd, int len)
+{
+	struct run_request *req = (struct run_request *)hd;
+	struct run *run;
+
+	run_request_in(req);
+
+	log_debug("receive_run_request %s from %d", req->uuid, hd->nodeid);
+
+	if (len != sizeof(struct run_request)) {
+		log_debug("receive_run_request %s bad len %d", req->uuid, len);
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	if (req->info.dest_nodeid && (req->info.dest_nodeid != our_nodeid))
+		return 0;
+
+	if (req->info.start_nodeid == our_nodeid) {
+		if (!(req->info.flags & DLMC_FLAG_RUN_START_NODE_RECV)) {
+			log_debug("receive_run_request ignore self");
+			return 0;
+		}
+
+		if (!opt(enable_helper_ind)) {
+			log_debug("receive_run_request %s helper not enabled", req->uuid);
+			return 0;
+		}
+
+		run = find_run(req->uuid);
+		if (!run) {
+			log_debug("receive_run_request from self no uuid %s", req->uuid);
+			return 0;
+		}
+
+		log_debug("receive_run_request %s to helper", req->uuid);
+
+		send_helper_run_request(req);
+		return 0;
+	}
+
+	if (!opt(enable_helper_ind)) {
+		log_debug("receive_run_request %s helper not enabled", req->uuid);
+		run->info.reply_count++;
+		run->info.need_replies--;
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	if (!(run = malloc(sizeof(struct run)))) {
+		log_error("receive_run_request %s no mem", req->uuid);
+		/* todo: send reply with failed */
+		return 0;
+	}
+
+	memset(run, 0, sizeof(struct run));
+
+	memcpy(run->uuid, req->uuid, RUN_UUID_LEN);
+	memcpy(run->command, req->command, RUN_COMMAND_LEN);
+	run->info.start_nodeid = req->info.start_nodeid;
+	run->info.dest_nodeid = req->info.dest_nodeid;
+	run->info.flags = req->info.flags;
+
+	list_add(&run->list, &run_ops);
+
+	log_error("run request %s %.128s", run->uuid, run->command);
+
+	log_debug("receive_run_request %s to helper", req->uuid);
+
+	send_helper_run_request(req);
+	/* todo: if no helper, send reply with failed */
+
+	return 0;
+}
+
+int send_run_request(struct run *run, struct run_request *req)
+{
+	struct node_daemon *node;
+	int i = 0;
+	int rv;
+
+	list_for_each_entry(node, &daemon_nodes, list) {
+		if (!node->daemon_member)
+			continue;
+
+		/*
+		 * When this starting node does not run the command,
+		 * there is no reply for our nodeid.
+		 */
+		if ((node->nodeid == our_nodeid) &&
+		    (run->info.flags & DLMC_FLAG_RUN_START_NODE_NONE))
+				continue;
+
+		/*
+		 * The command is only run on one specific node, and
+		 * only a reply from that node is needed.
+		 */
+		if (run->info.dest_nodeid && (node->nodeid != run->info.dest_nodeid))
+			continue;
+
+		run->node_count++;
+		run->node_results[i].nodeid = node->nodeid;
+		i++;
+	}
+
+	run->info.need_replies = run->node_count;
+
+	log_debug("send_run_request %s for %d nodes", req->uuid, run->node_count);
+
+	run_request_out(req);
+
+	rv = dlm_send_message_daemon((char *)req, sizeof(struct run_request));
+
+	return rv;
+}
+
+int send_run_reply(struct run *run, struct run_reply *rep)
+{
+	int rv;
+
+	log_debug("send_run_reply %s result %d", rep->uuid, rep->info.local_result);
+
+	run_reply_out(rep);
+
+	rv = dlm_send_message_daemon((char *)rep, sizeof(struct run_reply));
+
+	/*
+	 * If we are not the starting node, clear the run operation.
+	 */
+	if (rep->info.start_nodeid != our_nodeid)
+		clear_run(run);
+
+	return rv;
+}
+
 static void confchg_cb_daemon(cpg_handle_t handle,
 			      const struct cpg_name *group_name,
 			      const struct cpg_address *member_list,
